@@ -38,8 +38,85 @@
 static  XDeviceInfo  *devices = NULL;
 #endif
 
+static gchar *saved_input_device = NULL;
+
+static gchar *saved_corepointer_device = NULL;
+
 int gok_input_types[N_INPUT_TYPES];
 
+gboolean
+gok_input_detach_corepointer (void)
+{
+  int		num_devices, i;
+  Display      *display;
+  GtkWidget    *pMainWindow = gok_main_get_main_window ();
+#if 0
+#ifdef HAVE_XINPUT
+  if (pMainWindow && pMainWindow->window) 
+  {
+	  XDeviceInfo *pointer = NULL, *ext = NULL;
+	  display = GDK_WINDOW_XDISPLAY (pMainWindow->window); 
+	  if (devices) 
+	  {
+		XFreeDeviceList (devices);
+		devices = NULL;
+	  }
+ 
+	  devices = XListInputDevices(display, &num_devices);
+	  for (i = 0; i < num_devices; i++) 
+	  {
+		  if (!pointer && (devices[i].use == IsXPointer)) 
+		  {
+			  pointer = &devices[i];  
+		  }
+		  else if (!ext && (devices[i].use == IsXExtensionDevice))
+		  {
+			  XDevice *new_pointer = XOpenDevice (display, devices[i].id);
+			  int j;
+			  g_message ("new pointer %s: %d classes", devices[i].name, devices[i].num_classes);
+			  for (j = 0; j < devices[i].num_classes; j++) {
+				  if ((new_pointer->classes[j].input_class == ButtonClass) || 
+				      (new_pointer->classes[j].input_class == ValuatorClass))
+				  {
+					  ext = &devices[i];
+					  break;
+				  }
+			  }
+		  }
+	  }
+	  if (pointer && ext)
+	  {
+		  XChangePointerDevice (display, XOpenDevice (display, ext->id), 0, 1);
+		  XOpenDevice (display, pointer->id);
+		  g_message ("pointer swapped.");
+		  return TRUE;
+	  }
+	  else g_warning ("No extension devices or no corepointer device found.");
+  }
+#endif
+#endif
+  return FALSE;
+}
+
+gboolean   
+gok_input_ext_devices_exist (void)
+{
+	gboolean devices_exist = FALSE;
+	GSList *list = gok_input_get_device_list ();
+	if (list)
+	{
+		GSList *tmp = list;
+		while (tmp)
+		{
+			g_free (tmp->data);
+			tmp = tmp->next;
+		}
+		g_slist_free (list);
+		devices_exist = TRUE;
+	}
+	return devices_exist;
+}
+
 GSList *
 gok_input_get_device_list (void)
 {
@@ -137,8 +214,86 @@ find_input (Display	*display,
 	return input;
 }
 
+static GokInput*
+find_input_by_id (Display     *display,
+		  guint        id,
+		  gboolean     extended_only)
+{
+	int		i;
+	int		num_devices;
+	GokInput *input = NULL;
+	GokInput *temp;
+	
+	if (devices) {
+		XFreeDeviceList (devices);
+		devices = NULL;
+	}
+
+	devices = XListInputDevices(display, &num_devices);
+	
+	for(i = 0; i < num_devices; i++) {
+		if (extended_only)
+			if (devices[i].use != IsXExtensionDevice)
+				continue;
+
+		if ((guint) devices[i].id == id) {
+			temp = g_new0 (GokInput, 1);
+			temp->name = g_strdup (devices[i].name);
+			temp->info = &devices[i];
+
+			if (input == NULL)
+				input = temp;
+			else {
+				GokInput *temp1 = input;
+
+				while (temp1->next)
+					temp1 = temp1->next;	
+
+				temp1->next = temp;
+			}
+			
+		}
+	}
+	return input;
+}
+
 #endif
 
+static gboolean   
+gok_input_open (GokInput *input)
+{
+  GtkWidget *window = gok_main_get_main_window ();
+
+#ifdef HAVE_XINPUT  
+  if (window && window->window && 
+      input && input->info && (input->info->use == IsXExtensionDevice))
+    {
+      input->device = XOpenDevice(GDK_WINDOW_XDISPLAY (window->window), 
+				  input->info->id);
+      return TRUE;
+    }
+#endif
+  if (input )
+    g_warning ("could not open device %s", input);
+
+  return FALSE;
+}
+
+static void
+gok_input_close (GokInput *input)
+{
+  GtkWidget *window = gok_main_get_main_window ();
+
+#ifdef HAVE_XINPUT  
+  if (window && window->window && 
+      input && input->info && (input->info->use == IsXExtensionDevice))
+    {
+      XCloseDevice(GDK_WINDOW_XDISPLAY (window->window), 
+		   input->device);
+    }
+#endif
+}
+
 gboolean
 gok_input_init (GokInput *input, GdkFilterFunc filter_func)
 {
@@ -240,6 +395,23 @@ gok_input_free (GokInput *pInput)
 	g_free (pInput);
 }
 
+GokInput * 
+gok_input_find_by_device_id (guint id, gboolean extended_only)
+{
+  GtkWidget *window = gok_main_get_main_window ();
+  GokInput *input = NULL;
+#ifdef HAVE_XINPUT
+  
+  if (window && window->window) 
+    {
+      input = find_input_by_id (GDK_WINDOW_XDISPLAY (window->window), id, extended_only);
+    }
+#else
+  g_message ("Warning, XInput extension not present, cannot open input device.");
+#endif
+  return input;
+}
+
 GokInput * 
 gok_input_find_by_name (char *name, gboolean extended_only)
 {
@@ -257,24 +429,28 @@ gok_input_find_by_name (char *name, gboolean extended_only)
   return input;
 }
 
-gboolean   
-gok_input_open (GokInput *input)
+void
+gok_input_open_all (void)
 {
-  GtkWidget *window = gok_main_get_main_window ();
+	GSList *devices = gok_input_get_device_list ();
+	while (devices)
+	{
+		gok_input_open (devices->data);
+		devices = devices->next;
+	}
+}
 
-#ifdef HAVE_XINPUT  
-  if (window && window->window && 
-      input && input->info && (input->info->use == IsXExtensionDevice))
-    {
-      input->device = XOpenDevice(GDK_WINDOW_XDISPLAY (window->window), 
-				  input->info->id);
-      return TRUE;
-    }
-#endif
-  if (input )
-    g_warning ("could not open device %s", input);
-
-  return FALSE;
+void
+gok_input_close_others (GokInput *input)
+{
+	GSList *devices = gok_input_get_device_list ();
+	while (devices)
+	{
+		GokInput *tmp = devices->data;
+		if (tmp && input->device->device_id != tmp->device->device_id) 
+			gok_input_close (tmp);
+		devices = devices->next;
+	}
 }
 
 GokInput * gok_input_get_current (void)
@@ -287,13 +463,12 @@ GokInput * gok_input_get_current (void)
 
   /* was the input device specified on the command line? */
   input_device_name = gok_main_get_inputdevice_name ();
-  if (input_device_name == NULL) {
+  if ((input_device_name == NULL) &&  
+	  (gok_gconf_get_string (gconf_client_get_default (),
+       GOK_GCONF_INPUT_DEVICE,  &input_device_name))) {
 	  /* use gconf setting */
-	  gok_gconf_get_string (gconf_client_get_default (), 
-				GOK_GCONF_INPUT_DEVICE, 
-				&input_device_name);
-	  input = gok_input_find_by_name (input_device_name, TRUE);
-	  g_free (input_device_name);
+	input = gok_input_find_by_name (input_device_name, TRUE);
+	g_free (input_device_name);
   }
   else
   {
@@ -301,3 +476,23 @@ GokInput * gok_input_get_current (void)
   }
   return input;
 }
+
+void
+gok_input_restore (void)
+{
+	gok_input_set_extension_device_by_name (saved_input_device);
+	gok_input_open (gok_input_find_by_name (saved_input_device, TRUE));  
+}
+
+void
+gok_input_mismatch_warn (XDeviceInfo *device)
+{
+	gchar *message = g_strdup_printf (_("GOK has detected activity from a new hardware device named \'%s\'.  "
+					    "Would you like to use this device instead of device \'%s\'?"),
+					  device->name, gok_input_get_extension_device_name ());
+	saved_input_device = gok_input_get_extension_device_name ();
+	gok_input_set_extension_device_by_name (device->name);
+	gok_main_warn (message, TRUE, gok_input_restore);
+	g_free (message);
+}
+
