@@ -698,6 +698,11 @@ gok_spy_get_list_parent (Accessible *search_root)
 		    gok_spy_accessible_unref (child);
 		    return search_root;
 		}
+		else if (role == SPI_ROLE_LIST)
+		{
+		    gok_spy_accessible_unref (search_root);
+		    return child;
+		}
 		else 
 		{
 		    gok_spy_accessible_unref (search_root);
@@ -1392,12 +1397,13 @@ gok_spy_resolve_namesakes (GSList *list)
 	/* for every node in the given list... */
 	while (list && list->next) {
 		AccessibleNode *node = list->data;
+		GSList *lp = list;
 		if (node && node->pname) 
 		{
 		    strncpy (name, node->pname, 63);
 		}
 		/* for every node with the same name as the current node... */
-		while (list && (namesake = gok_spy_node_get_node_named ( list->next, name )) 
+		while (lp && (namesake = gok_spy_node_get_node_named ( lp->next, name )) 
 				!= NULL) 
 		{
 			if (!distinguished_source) {
@@ -1405,7 +1411,7 @@ gok_spy_resolve_namesakes (GSList *list)
 				gok_spy_distinguish_node_name ( (AccessibleNode *) list->data );
 			}
 			gok_spy_distinguish_node_name ( (AccessibleNode *) namesake->data );
-			list = list->next;
+			lp = lp->next;
 		}
 		distinguished_source = FALSE;	
 		if (list) list = list->next;
@@ -1682,16 +1688,17 @@ GSList* gok_spy_append_node (GSList* nodes,
 			if (caption) 
 			{
 				char *caption_string = Accessible_getName (caption);
-				if (caption_string) 
+				if (caption_string && strlen (caption_string)) 
 				{
 					pName = g_strndup (caption_string, 20);
+				        gok_log ("Table caption: %s", caption_string);
 					SPI_freeString (caption_string);
 				}
 			}
 			if (pName == NULL && AccessibleTable_getNColumns (table) == 1) 
 			{
 				char *desc = AccessibleTable_getColumnDescription (table, 0);
-				if (desc) 
+				if (desc && strlen (desc)) 
 				{
 					pName = g_strndup (desc, 20);
 					SPI_freeString (desc);
@@ -1703,8 +1710,9 @@ GSList* gok_spy_append_node (GSList* nodes,
 					if (header) 
 					{
 						desc = Accessible_getName (header);
-						if (desc) 
+						if (desc && strlen (desc)) 
 						{
+						        gok_log ("Table col header: %s", desc);
 							pName = g_strndup (desc, 20);
 							SPI_freeString (desc);
 						}
@@ -1722,6 +1730,8 @@ GSList* gok_spy_append_node (GSList* nodes,
 		if (pName != NULL) {
 			gok_log("we have a name: [%s]",pName);
 			/* create a new node */
+			if (*g_strstrip (pName) == 0 || strlen (pName) == 0) 
+			    g_warning ("EMPTY Node name: [%s]", pName);
 			nodes = gok_spy_add_node (nodes, pAccessible, flags, 0, pName);
 		}
 		/* if we decide to allow access to nameless menus ...
@@ -1788,6 +1798,8 @@ GSList* gok_spy_append_node (GSList* nodes,
 				}
 			}
 			if (pName != NULL) {
+			        if (*g_strstrip (pName) == 0 || strlen (pName) == 0) 
+				    g_warning ("EMPTY Node name: [%s]", pName);
 				gok_log ("we have a name: [%s]",pName);
 				nodes = gok_spy_add_node (nodes, pAccessible, flags, 0, pName);
 			}
@@ -2276,7 +2288,11 @@ gok_spy_find_and_append_toolbar_items (Accessible *root, AccessibleNodeFlags fla
 	gint i, nchildren;
 	gboolean retval = FALSE;
 
-	if (Accessible_isAction (root)) {
+	if (gok_spy_is_menu_role (Accessible_getRole (root)))
+	{
+	        return FALSE; /* make sure we continue searching independent of toolbars here */
+	}
+	else if (Accessible_isAction (root)) {
 		_priv_ui_nodes = gok_spy_append_node (_priv_ui_nodes, root, flags);
 		retval = TRUE;
 		/* uncomment if we don't care about children of actionables: */
@@ -2376,8 +2392,8 @@ gok_spy_search_component_list (Accessible *rootAccessible, GokSpyUIFlags keyboar
 				if (non_context_flags.value) 
 				{
 				    _priv_ui_nodes = gok_spy_append_node (_priv_ui_nodes, child, flags);
-				    /* special case for notebook tabs, whose children should also appear */
-  				    if (role == SPI_ROLE_PAGE_TAB)
+				    /* special case for notebook tabs and expanders, whose children should also appear */
+  				    if (role == SPI_ROLE_PAGE_TAB || role == SPI_ROLE_TOGGLE_BUTTON)
   				    {
   					ui_flags.value |= gok_spy_search_component_list (child, keyboard_ui_flags, context_flags).value;
   				    }
