@@ -46,6 +46,7 @@ typedef struct
 {
 	gchar *string;
 	gint   priority;
+	gboolean in_primary;
 } WordPrediction;
 
 /* The filename of the dictionary file */
@@ -54,13 +55,18 @@ static gchar *dictionary_full_path;
 /* declarations of implementation methods */
 static gchar** utf8_wordcomplete_predict_string (GokWordComplete *complete, const gchar* pWord, gint num_predictions);
 static gboolean utf8_wordcomplete_open (GokWordComplete *complete, const gchar *directory);
+static void utf8_wordcomplete_close (GokWordComplete *complete);
+static gboolean utf8_wordcomplete_add_new_word (GokWordComplete *complete, const gchar *word);
+static gboolean utf8_wordcomplete_increment_word_frequency (GokWordComplete *complete, const gchar *word);
+static gboolean utf8_wordcomplete_validate_word (GokWordComplete *complete, const gchar *word);
 
 /* private internal methods */
-static GSList *utf8_wordcomplete_find (GokUTF8WordComplete *complete, const gchar *word);
+static GList *utf8_wordcomplete_find (GokUTF8WordComplete *complete, const gchar *word);
+static GList *utf8_wordcomplete_find_word (GokUTF8WordComplete *complete, const gchar *word);
 static gint utf8_sort_by_priority (gconstpointer a, gconstpointer b, gpointer data);
 static gint utf8_sort_by_collation (gconstpointer a, gconstpointer b);
-static GSList *utf8_add_from_system_dicts (GSList *dictionary, const gchar *system_dict_filename);
-static GSList *utf8_add_to_list_from_lines (GSList *list, gchar **lines, gboolean read_freqs);
+static GList *utf8_add_from_system_dicts (GList *dictionary, const gchar *system_dict_filename);
+static GList *utf8_add_to_list_from_lines (GList *list, gchar **lines, gboolean read_freqs, gboolean in_primary);
 static void utf8_wordcomplete_create_unicode_hash (GokUTF8WordComplete *complete);
 static void utf8_apply_case (const gchar *word,  gchar **predictions);
 
@@ -86,11 +92,13 @@ gok_utf8_wordcomplete_class_init (GokUTF8WordCompleteClass *klass)
 {
 	GokWordCompleteClass *word_complete_class = (GokWordCompleteClass *) klass;
 	word_complete_class->open = utf8_wordcomplete_open;
+	word_complete_class->close = utf8_wordcomplete_close;
 	word_complete_class->predict_string = utf8_wordcomplete_predict_string;
+	word_complete_class->add_new_word = utf8_wordcomplete_add_new_word;
+	word_complete_class->increment_word_frequency = utf8_wordcomplete_increment_word_frequency;
+	word_complete_class->validate_word = utf8_wordcomplete_validate_word;
 }
 
-/* TODO FIXME need _close, to deallocate the hashtable and list(s) */
-
 static gboolean
 utf8_wordcomplete_open (GokWordComplete *complete, const gchar *directory)
 {
@@ -109,9 +117,15 @@ utf8_wordcomplete_open (GokWordComplete *complete, const gchar *directory)
 
 	if (!g_file_get_contents (dictionary_full_path, &contents, NULL, &error))
 	{
-		fprintf (stderr, "could not get file contents\n");
+	        g_warning (_("Could not read contents of dictionary file \'%s\'\n"), 
+			   dictionary_full_path);
 		return FALSE;
 	}
+	else
+	{
+		GOK_UTF8WORDCOMPLETE (complete)->primary_dict_filename = g_strdup (dictionary_full_path);
+		g_free (dictionary_full_path);
+	}
 
 	/* split into lines */
 	dictionary_lines = 
@@ -124,14 +138,13 @@ utf8_wordcomplete_open (GokWordComplete *complete, const gchar *directory)
 		return FALSE;
 	}
 
-
 	/* ignore first line, it's just "WPDictFile" */
 	if (!strcmp (dictionary_lines[0], "WPDictFile"))
 	{
 		if (dictionary_lines[1]) /* sanity check, make sure there's at least one word! */
 		{
 			utf8_complete->word_list = utf8_add_to_list_from_lines (utf8_complete->word_list, 
-										&dictionary_lines[1], TRUE);
+										&dictionary_lines[1], TRUE, TRUE);
 		}
 	}
 
@@ -144,27 +157,66 @@ utf8_wordcomplete_open (GokWordComplete *complete, const gchar *directory)
 	}
 
 	/* sort the list, so we don't have to search exhaustively */
-	utf8_complete->word_list = g_slist_sort (utf8_complete->word_list, 
+	utf8_complete->word_list = g_list_sort (utf8_complete->word_list, 
 						 utf8_sort_by_collation);
 
-	utf8_complete->word_list_end = g_slist_last (utf8_complete->word_list);
+	utf8_complete->word_list_end = g_list_last (utf8_complete->word_list);
 
 	/* create gunichar hash for looking up search entry points */
 	utf8_wordcomplete_create_unicode_hash (utf8_complete);
 
 	fprintf (stderr, "Word prediction dictionary contains a total of %d words\n", 
-		 g_slist_length (utf8_complete->word_list));
+		 g_list_length (utf8_complete->word_list));
 
 	return TRUE;
 }
 
+static void
+utf8_wordcomplete_write_line (gpointer data, gpointer user_data)
+{
+	GIOChannel *io = user_data;
+	GError *error = NULL;
+	gsize bytes;
+	WordPrediction *word_info = data;
+	gchar line[256];
+	snprintf (line, 255, "%s\t%d\t%d\n", word_info->string, word_info->priority, 2);
+	g_io_channel_write_chars (io, line, -1, &bytes, &error);
+}
+
+static void
+utf8_wordcomplete_close (GokWordComplete *complete)
+{
+	GIOChannel *io;
+	GError *error = NULL;
+	gsize bytes;
+	GokUTF8WordComplete *ucomplete = GOK_UTF8WORDCOMPLETE (complete);
+	/* open for writing */
+	io = g_io_channel_new_file (ucomplete->primary_dict_filename, "w", &error);
+	/* write the header */
+	if (!error)
+		g_io_channel_write_chars (io, "WPDictFile\n", -1, &bytes, &error);
+	if (!error)
+	{
+		/* write entries*/
+		g_list_foreach (ucomplete->word_list, utf8_wordcomplete_write_line, io);
+	}
+	g_list_free (ucomplete->word_list);
+	g_list_free (ucomplete->word_list_end);
+	g_list_free (ucomplete->start_search);
+	g_list_free (ucomplete->end_search);
+	g_hash_table_destroy (ucomplete->unicode_start_hash);
+	g_hash_table_destroy (ucomplete->unicode_end_hash);
+	g_free (ucomplete->primary_dict_filename);
+	g_io_channel_shutdown (io, TRUE, &error);
+}
+
 static gchar**
 utf8_wordcomplete_predict_string (GokWordComplete *complete, const gchar* word, gint num_predictions)
 {
 	gchar **word_predict_list;
 	gchar *normalized, *tmp;
 	gint  i, n, m = num_predictions;
-	GSList *predictions;
+	GList *predictions;
 	GokUTF8WordComplete *utf8_complete = GOK_UTF8WORDCOMPLETE (complete);
 	gpointer data;
 	
@@ -178,7 +230,7 @@ utf8_wordcomplete_predict_string (GokWordComplete *complete, const gchar* word,
 	   reset the search pointers if need be */
 	if (!utf8_complete->most_recent_word || !g_str_has_prefix (word, utf8_complete->most_recent_word))
 	{
-		GSList *start, *end;
+		GList *start, *end;
 		gunichar unicode_char = g_utf8_get_char (word);
 		if (utf8_complete->unicode_start_hash)
 			start = g_hash_table_lookup (utf8_complete->unicode_start_hash, (gpointer) unicode_char);
@@ -210,7 +262,7 @@ utf8_wordcomplete_predict_string (GokWordComplete *complete, const gchar* word,
 	predictions = utf8_wordcomplete_find (utf8_complete, normalized);
 
 	/* reorder according to priority */
-	predictions = g_slist_sort_with_data (predictions, utf8_sort_by_priority, complete); 
+	predictions = g_list_sort_with_data (predictions, utf8_sort_by_priority, complete); 
 
 	/* convert to array of strings */
 	n = -1;
@@ -218,14 +270,14 @@ utf8_wordcomplete_predict_string (GokWordComplete *complete, const gchar* word,
 	{
 		WordPrediction *prediction;
 		++n;
-		prediction = g_slist_nth_data (predictions, n);
+		prediction = g_list_nth_data (predictions, n);
 		word_predict_list [n] = prediction ? g_strdup (prediction->string) : NULL;
 	} while (word_predict_list [n] && n < num_predictions);
 
 	word_predict_list [num_predictions] = NULL; /* in case predictions-length >= num_predictions */
 
 	/* free the extra storage */
-	g_slist_free (predictions);
+	g_list_free (predictions);
 
 	/* were there any predictions? */
 	if (n == 0)
@@ -245,15 +297,54 @@ utf8_wordcomplete_predict_string (GokWordComplete *complete, const gchar* word,
 }
 
 static WordPrediction *
-utf8_word_prediction_new (const gchar *word, gint priority)
+utf8_word_prediction_new (const gchar *word, gint priority, gboolean in_primary)
 {
 	WordPrediction *prediction = g_new0 (WordPrediction, 1);
 	prediction->string = word ? g_utf8_casefold (word, -1): NULL;
 	prediction->priority = priority;
+	prediction->in_primary = in_primary;
 	g_assert (prediction->string);
 	return prediction;
 }
 
+static gboolean 
+utf8_wordcomplete_add_new_word (GokWordComplete *complete, const gchar *word)
+{
+	WordPrediction *prediction = utf8_word_prediction_new (word, 1, TRUE);
+	GokUTF8WordComplete *ucomplete = GOK_UTF8WORDCOMPLETE (complete);
+	ucomplete->word_list = g_list_sort (g_list_prepend (ucomplete->word_list, prediction),
+					    utf8_sort_by_collation);
+	ucomplete->word_list_end = g_list_last (ucomplete->word_list);
+	utf8_wordcomplete_create_unicode_hash (ucomplete);
+	return TRUE;
+}
+
+static gboolean 
+utf8_wordcomplete_increment_word_frequency (GokWordComplete *complete, const gchar *word)
+{
+	WordPrediction *prediction;
+	GokUTF8WordComplete *ucomplete = GOK_UTF8WORDCOMPLETE (complete);
+	GList *node = utf8_wordcomplete_find_word (ucomplete, word);
+	if (node)
+	{
+		WordPrediction *prediction = node->data;
+		if (prediction)
+		{
+			prediction->priority++;
+			return TRUE;
+		}
+	}
+	return FALSE;
+}
+
+static gboolean 
+utf8_wordcomplete_validate_word (GokWordComplete *complete, const gchar *word)
+{
+	GokUTF8WordComplete *ucomplete = GOK_UTF8WORDCOMPLETE (complete);
+	GList *node = utf8_wordcomplete_find_word (ucomplete, word);
+	return node ? TRUE : FALSE;
+}
+
 static WordPredictionCaseType
 utf8_determine_case (const gchar *word)
 {
@@ -389,10 +480,40 @@ utf8_apply_case (const gchar *word,  gchar **predictions)
 	}
 }
 
-static GSList *
+static GList *
+utf8_wordcomplete_find_word (GokUTF8WordComplete *complete, const gchar *word)
+{
+	GList *next = complete->start_search;
+	WordPrediction *prediction;
+
+	if (next && next->prev) 
+	{
+	    next = next->prev;
+	}
+	if (word) 
+	{
+		while (next && next->data) 
+		{
+			prediction = next->data;
+			if (prediction->string && !strcmp (prediction->string, word))
+			{
+				return next;
+			}
+			else if (prediction->string && !g_str_has_prefix (prediction->string, word))
+			{
+				return NULL;
+			}
+			next = g_list_next (next);
+		}
+	}
+
+	return NULL;
+}
+
+static GList *
 utf8_wordcomplete_find (GokUTF8WordComplete *complete, const gchar *word)
 {
-	GSList *found = NULL, *next = complete->start_search;
+	GList *found = NULL, *next = complete->start_search;
 	WordPrediction *prediction;
 
 	if (word) 
@@ -405,7 +526,7 @@ utf8_wordcomplete_find (GokUTF8WordComplete *complete, const gchar *word)
 			    g_utf8_strlen (prediction->string, -1) > word_len) 
 			{
 				if (!found) complete->start_search = next;
-				found = g_slist_append (found, prediction);
+				found = g_list_append (found, prediction);
 			}
 			else if (found) 
 			{
@@ -416,7 +537,7 @@ utf8_wordcomplete_find (GokUTF8WordComplete *complete, const gchar *word)
 			{
 				break;
 			}
-			next = g_slist_next (next);
+			next = g_list_next (next);
 		}
 	}
 
@@ -448,8 +569,8 @@ utf8_sort_by_collation (gconstpointer a, gconstpointer b)
 		return strcmp (p_a->string, p_b->string);
 }
 
-static GSList *
-utf8_add_from_system_dicts (GSList *dictionary, const gchar *system_dict_filenames)
+static GList *
+utf8_add_from_system_dicts (GList *dictionary, const gchar *system_dict_filenames)
 {
 	gchar *contents;
 	gchar **system_dict_lines = NULL;
@@ -478,7 +599,7 @@ utf8_add_from_system_dicts (GSList *dictionary, const gchar *system_dict_filenam
 			
 			if (system_dict_lines != NULL) 
 			{
-				dictionary = utf8_add_to_list_from_lines (dictionary, system_dict_lines, FALSE);
+				dictionary = utf8_add_to_list_from_lines (dictionary, system_dict_lines, FALSE, FALSE);
 				g_strfreev (system_dict_lines);
 			}
 			++i;
@@ -488,8 +609,8 @@ utf8_add_from_system_dicts (GSList *dictionary, const gchar *system_dict_filenam
 	return dictionary;
 }
 
-static GSList *
-utf8_add_to_list_from_lines (GSList *list, gchar **lines, gboolean read_freqs)
+static GList *
+utf8_add_to_list_from_lines (GList *list, gchar **lines, gboolean read_freqs, gboolean in_primary)
 {
 	while (*lines)
 	{
@@ -532,9 +653,9 @@ utf8_add_to_list_from_lines (GSList *list, gchar **lines, gboolean read_freqs)
 				string = g_utf8_normalize (string, -1, G_NORMALIZE_DEFAULT_COMPOSE); /* this allocs a new string */
 				priority = (read_freqs && tokens && tokens[0] && tokens[1]) ? 
 					g_ascii_strtod (tokens[1], NULL) : DEFAULT_INITIAL_FREQUENCY;
-				prediction = utf8_word_prediction_new (string, priority);
+				prediction = utf8_word_prediction_new (string, priority, in_primary);
 				/* prepend is much faster than append */
-				list = g_slist_prepend (list, prediction);
+				list = g_list_prepend (list, prediction);
 			}
 			else
 			{
@@ -552,7 +673,7 @@ utf8_add_to_list_from_lines (GSList *list, gchar **lines, gboolean read_freqs)
 static void
 utf8_wordcomplete_create_unicode_hash (GokUTF8WordComplete *complete)
 {
-	GSList *next = complete->word_list;
+	GList *next = complete->word_list;
 
 	if (complete->unicode_start_hash) 
 		g_hash_table_destroy (complete->unicode_start_hash);
